# /**
#  * @author [Emmanuel Castillo]
#  * @email [excastillot@unal.edu.co]
#  * @create date 2021-10-12 10:58:25
#  * @modify date 2021-10-12 10:58:25
#  * @desc [description]
#  */
######## 
# Moment magnitude is adapted from 
# https://github.com/krischer/moment_magnitude_calculator/tree/master/scripts
# :copyright:
# Lion Krischer (krischer@geophysik.uni-muenchen.de), 2012

# import colorama
import matplotlib.pylab as plt
# import mtspec
import numpy as np
# from obspy.core.event import Magnitude as Mag, catalog
from obspy.core.event import Magnitude as Mag
from obspy.core.event import Catalog
from obspy.core.event import read_events, Comment
from obspy.io.xseed.parser import Parser
from obspy.core.event import WaveformStreamID
# import progressbar
import scipy
import scipy.optimize
# import warnings
from obspy import UTCDateTime
from obspy.clients.filesystem.sds import Client
from obspy.io.xseed.parser import Parser
from obspy.geodetics import gps2dist_azimuth
from obspy.signal.invsim import estimate_magnitude
from math import log10
import concurrent.futures as cf
import numpy as np
import time
import json

paz_wa = {'sensitivity': 2800, 'zeros': [0j], 'gain': 1,
          'poles': [-6.2832 - 4.7124j, -6.2832 + 4.7124j]}

def fit_spectrum(spectrum, frequencies, traveltime, initial_omega_0,
    initial_f_c,quality_factor=10000):
    """
    Fit a theoretical source spectrum to a measured source spectrum.

    Uses a Levenburg-Marquardt algorithm.

    :param spectrum: The measured source spectrum.
    :param frequencies: The corresponding frequencies.
    :para traveltime: Event traveltime in [s].
    :param initial_omega_0: Initial guess for Omega_0.
    :param initial_f_c: Initial guess for the corner frequency.


    :returns: Best fits and standard deviations.
        (Omega_0, f_c, Omega_0_std, f_c_std)
        Returns None, if the fit failed.
    """
    def f(frequencies, omega_0, f_c):
        return calculate_source_spectrum(frequencies, omega_0, f_c,
                quality_factor, traveltime)
    popt, pcov = scipy.optimize.curve_fit(f, frequencies, spectrum, \
        p0=list([initial_omega_0, initial_f_c]), maxfev=100000)
    if popt is None:
        return None
    return popt[0], popt[1], pcov[0, 0], pcov[1, 1]

def calculate_source_spectrum(frequencies, omega_0, corner_frequency, Q,
    traveltime):
    """
    After Abercrombie (1995) and Boatwright (1980).

    Abercrombie, R. E. (1995). Earthquake locations using single-station deep
    borehole recordings: Implications for microseismicity on the San Andreas
    fault in southern California. Journal of Geophysical Research, 100,
    24003–24013.

    Boatwright, J. (1980). A spectral theory for circular seismic sources,
    simple estimates of source dimension, dynamic stress drop, and radiated
    energy. Bulletin of the Seismological Society of America, 70(1).

    The used formula is:
        Omega(f) = (Omege(0) * e^(-pi * f * T / Q)) / (1 + (f/f_c)^4) ^ 0.5

    :param frequencies: Input array to perform the calculation on.
    :param omega_0: Low frequency amplitude in [meter x second].
    :param corner_frequency: Corner frequency in [Hz].
    :param Q: Quality factor.
    :param traveltime: Traveltime in [s].
    """
    num = omega_0 * np.exp(-np.pi * frequencies * traveltime / Q)
    denom = (1 + (frequencies / corner_frequency) ** 4) ** 0.5
    return num / denom

def fit_moment_magnitude_relation_curve(Mls, Mws, Mw_stds):
    """
    Fits a quadratic curve to
        Mw = a + b * Ml + c * Ml ** 2
    Returns the best fitting [a, b, c]
    """
    def y(x, a, b, c):
        return a + b * x + c * x ** 2
    # Use a straight line as starting point.
    Mls = np.ma.masked_invalid(Mls)
    Mws = np.ma.masked_invalid(Mws)
    inds = ~(Mls.mask | Mws.mask | np.isnan(Mw_stds) | (Mw_stds <= 0))
    popt, pcov = scipy.optimize.curve_fit(y, Mls[inds], Mws[inds], \
        p0=[0.0, 1.0, 0.0], sigma=Mw_stds[inds], maxfev=100000)
    return popt[0], popt[1], popt[2]

def plot_ml_vs_mw(catalog):
    moment_magnitudes = []
    moment_magnitudes_std = []
    local_magnitudes = []
    local_magnitudes_std = []
    for event in catalog:
        Mw = None
        Mw_std = None
        Ml = None
        Ml_std = None
        for mag in event.magnitudes:
            if Mw is not None and Ml is not None:
                break
            mag_type = mag.magnitude_type.lower()
            if mag_type == "mw":
                if Mw is not None:
                    continue
                Mw = mag.mag
                Mw_std = mag.mag_errors.uncertainty
            elif mag_type == "ml":
                if Ml is not None:
                    continue
                Ml = mag.mag
                Ml_std = mag.mag_errors.uncertainty
        moment_magnitudes.append(Mw)
        moment_magnitudes_std.append(Mw_std)
        local_magnitudes.append(Ml)
        local_magnitudes_std.append(Ml_std)
    moment_magnitudes = np.array(moment_magnitudes, dtype="float64")
    moment_magnitudes_std = np.array(moment_magnitudes_std, dtype="float64")
    local_magnitudes = np.array(local_magnitudes, dtype="float64")
    local_magnitudes_std = np.array(local_magnitudes_std, dtype="float64")

    # Fit a curve through the data.
    a, b, c = fit_moment_magnitude_relation_curve(local_magnitudes,
            moment_magnitudes, moment_magnitudes_std)
    x_values = np.linspace(-2.0, 4.0, 10000)
    fit_curve = a + b * x_values + c * x_values ** 2

    plt.figure(figsize=(10, 8))
    # Show the data values as dots.
    plt.scatter(local_magnitudes, moment_magnitudes, color="blue",
        edgecolor="black")

    # Plot the Ml=Mw line.
    plt.plot(x_values, x_values, label="$Mw=Ml$", color="k", alpha=0.8)
    plt.plot(x_values, 0.67 + 0.56 * x_values + 0.046 * x_values ** 2,
        label="$Mw=0.67 + 0.56Ml + 0.046Ml^2 (gruenthal etal 2003)$", color="green", ls="--")
    plt.plot(x_values, 0.53 + 0.646 * x_values + 0.0376 * x_values ** 2,
        label="$Mw=0.53 + 0.646Ml + 0.0376Ml^2 (gruenthal etal 2009)$", color="green")
    plt.plot(x_values, 0.594 * x_values + 0.985,
        label="$Mw=0.985 + 0.594Ml (goertz-allmann etal 2011)$", color="orange")
    plt.plot(x_values, (x_values + 1.21) / 1.58,
        label="$Mw=(Ml + 1.21) / 1.58 (bethmann etal 2011)$", color="red")
    plt.plot(x_values, fit_curve, color="blue",
        label="$Data$ $fit$ $with$ $Mw=%.2f + %.2fMl + %.3fMl^2$" % (a, b, c))
    # Set limits and labels.
    plt.xlim(-2, 4)
    plt.ylim(-2, 4)
    plt.xlabel("Ml", fontsize="x-large")
    plt.ylabel("Mw", fontsize="x-large")
    # Show grid and legend.
    plt.grid()
    plt.legend(loc="lower right")
    plt.savefig("moment_mag_automatic.pdf")

def plot_source_radius(cat):
    mw = []
    mw_std = []
    source_radius = []
    source_radius_std = []

    plt.figure(figsize=(10, 4.5))

    # Read the source radius.
    for event in cat:
        mag = event.magnitudes[1]
        if len(mag.comments) != 2:
            continue
        mw.append(mag.mag)
        mw_std.append(mag.mag_errors.uncertainty)
        sr, std = mag.comments[1].text.split(";")
        _, sr = sr.split("=")
        _, std = std.split("=")
        sr = float(sr[:-1])
        std = float(std)
        source_radius.append(sr)
        source_radius_std.append(std)
    plt.errorbar(mw, source_radius, yerr=source_radius_std,
        fmt="o", linestyle="None")
    plt.xlabel("Mw", fontsize="x-large")
    plt.ylabel("Source Radius [m]", fontsize="x-large")
    plt.grid()
    plt.savefig("/Users/lion/Desktop/SourceRadius.pdf")

class Magnitude():
    def __init__(self,client,cat,resp_file,
                jsonfile=None) -> None:
        self.client = client
        # super().__init__(sds_archive,sds_type,format)
        self.resp_file = resp_file
        self.cat = cat
        self.jsonfile = jsonfile
    
    @property
    def info_from_json(self):
        if self.jsonfile == None:
            return None
        else:
            f = open (self.jsonfile, "r")
            return json.load(f)

    @property
    def catalog(self):
        if isinstance(self.cat,Catalog):
            return self.cat
        else:
            return read_events(self.cat)

    @property
    def resp(self):
        return Parser(self.resp_file)

    def _get_channel_resp(self):
        parser = self.resp
        channels = [c['channel_id'] for c in parser.get_inventory()['channels']]
        # print(channels)
        # exit()
        parsers = dict.fromkeys(channels, parser)
        print(parsers)
        exit()
        return parsers

    def _get_paz_wa(self,seed_id):
        try:
            parse = self.resp
        except Exception as e:
            print(e)
        
        return parse.get_paz(seed_id)
    
    def _get_coordinates(self,seed_id):
        try:
            parse = self.resp
        except Exception as e:
            print(e)
        
        return parse.get_coordinates(seed_id)

    def _estimate_local_CC_magnitude(self,event_coords:tuple,times2trim:dict,
                            origin_time:UTCDateTime):
        event_lat,event_lon,event_depth = event_coords

        mags = []
        def compute_ml(t2t):
        # for info,times in times2trim.items():
            info,times = t2t
            net,sta,loc,cha = info.split(".")
            starttime,endtime = times
            try: 
                st = self.client.get_waveforms(net,sta,loc,cha, starttime,
                        endtime, metadata=True)
                assert(len(st) == 3)
            except Exception:
                print(info, "---")
                return

            amps = []
            pazs = []
            for tr in st:
                single_cha = ".".join((tr.stats.network,
                                        tr.stats.station,
                                        tr.stats.location,
                                        tr.stats.channel ))
                paz = self._get_paz_wa(single_cha)
                coords = self._get_coordinates(single_cha)

                # tr.simulate(paz_remove=paz, paz_simulate=paz_wa, 
                #             water_level=10)
                if tr.stats.channel[-1] != "Z":
                    amp =  max(abs(tr.data))
                    amps.append(amp)
                    pazs.append(paz)

            epi_dist, az, baz = gps2dist_azimuth(event_lat, event_lon,
                                             coords["latitude"], coords["longitude"])
            epi_dist = epi_dist / 1000

            timespan = endtime-starttime
            ml = estimate_magnitude(pazs,amps,[timespan,timespan],epi_dist)

            # print(sta, ml)
            mags.append(ml)
        
        with cf.ThreadPoolExecutor() as executor:
            executor.map(compute_ml,times2trim.items())

        net_mag = np.median(mags)
        print(f"{origin_time}: ({event_lat},{event_lon},{event_depth}) | Ml:{round(net_mag,3)}")
        return round(net_mag,3)

    def _estimate_local_SED_magnitude(self,event_coords:tuple,times2trim:dict,waterlevel=10):
        event_lat,event_lon = event_coords

        mags = []
        for info,times in times2trim.items():
            net,sta,loc,cha = info.split(".")
            starttime,endtime = times
            try: 
                st = self.client.get_waveforms(net,sta,loc,cha, starttime,
                        endtime, metadata=True)
                assert(len(st) == 3)
            except Exception:
                print(info, "---")
                continue

            amps = []
            for tr in st:
                single_cha = ".".join((tr.stats.network,
                                        tr.stats.station,
                                        tr.stats.location,
                                        tr.stats.channel ))
                paz = self._get_paz_wa(single_cha)
                coords = self._get_coordinates(single_cha)

                tr.simulate(paz_remove=paz, paz_simulate=paz_wa, 
                            water_level=waterlevel)
                if tr.stats.channel[-1] != "Z":
                    amp =  max(abs(tr.data))
                    amps.append(amp)

            ampl = max(amps)

            epi_dist, az, baz = gps2dist_azimuth(event_lat, event_lon,
                                             coords["latitude"], coords["longitude"])
            epi_dist = epi_dist / 1000


            a = 0.018
            b = 2.17
            ml = log10(ampl * 1000) + a * epi_dist + b
            print(sta, ml)
            mags.append(ml)
        
        net_mag = np.median(mags)
        print("Network magnitude:", net_mag)
        return net_mag

    def _estimate_moment_magnitudes(self, outfile=None,
                                vp=4266.96,vsp_factor=1.84,
                                density=2650.0,
                                time_before_pick = 0.2,
                                time_after_pick = 0.8,
                                padding = 20,waterlevel=10,
                                out_format="QUAKEML"):
        """
        :param cat: obspy.core.event.Catalog object.
        :param outfile: str

        # # Rock density in km/m^3.
        # # Velocities in m/s.

        # # time_before/after_pick: How many seconds before and after the pick
        #  to choose for calculating the spectra.

        # # Fixed quality factor. Very unstable inversion for it. Has almost no influence
        # # on the final seismic moment estimations but has some influence on the corner
        # # frequency estimation and therefore on the source radius estimation.
        """
        cat = self.catalog
        vs = vp/vsp_factor
        Mws = []
        Mls = []
        Mws_std = []
        print("wait",catalog)

        for event in cat:
            print(event)
            if not event.origins:
                print ("No origin for event %s" % event.resource_id)
                continue
            # if not event.magnitudes:
            #     print ("No magnitude for event %s" % event.resource_id)
            #     continue

            origin_time = event.origins[0].time
            latitude = event.origins[0].latitude
            longitude = event.origins[0].longitude
            depth = event.origins[0].depth
            # local_magnitude = event.magnitudes[0].mag

            #if local_magnitude < 1.0:
                #continue
            moments = []
            source_radii = []
            corner_frequencies = []
            # print(event.picks)
            # exit()
            for pick in event.picks:
                # Only p phase picks.
                if pick.phase_hint.lower() == "p":
                    radiation_pattern = 0.52
                    velocity = vp
                    k = 0.32
                elif pick.phase_hint.lower() == "s":
                    radiation_pattern = 0.63
                    velocity = vs
                    k = 0.21
                else:
                    continue
                distance = (pick.time - origin_time) * velocity
                if distance <= 0.0:
                    continue

                # wav_id = pick.waveform_id
                if self.jsonfile != None:
                    wav_id = pick.waveform_id
                    sta = wav_id.station_code
                    net = self.info_from_json[sta]["network"]
                    cha = self.info_from_json[sta]["channels"][0][:2]+"*"
                    pick.waveform_id = WaveformStreamID(network_code=net,
                                                    station_code=sta,
                                                    location_code="*",
                                                    channel_code=cha)
                # print(pick.waveform_id, pick.time,
                #                                 padding,waterlevel)

                stream = self.get_corresponding_stream(pick.waveform_id, pick.time,
                                                padding,waterlevel)
                # exit()
                # print(pick.waveform_id)
                # print(stream)
                if stream is None or len(stream) != 3:
                    continue
                omegas = []
                corner_freqs = []
                for trace in stream:
                    # print(pick.time)
                    # print(trace.stats.starttime)
                    # print(trace.stats.delta)
                    # Get the index of the pick.
                    pick_index = int(round((pick.time - trace.stats.starttime) / \
                        trace.stats.delta))
                    # print(len(trace.data))
                    # print(pick_index - \
                    #     int(time_before_pick * trace.stats.sampling_rate))
                    # print(pick_index + int(time_after_pick * trace.stats.sampling_rate))
                    # Choose date window 0.5 seconds before and 1 second after pick.
                    print(trace.data)
                    data_window = trace.data[pick_index - \
                        int(time_before_pick * trace.stats.sampling_rate): \
                        pick_index + int(time_after_pick * trace.stats.sampling_rate)]
                    # print(data_window)
                    # Calculate the spectrum.
                    spec, freq = mtspec.mtspec(data_window, trace.stats.delta, 2)
                    try:
                        fit = fit_spectrum(spec, freq, pick.time - origin_time,
                                spec.max(), 10.0)
                    except:
                        continue
                    if fit is None:
                        continue
                    # print(fit)
                    Omega_0, f_c, err, _ = fit
                    Omega_0 = np.sqrt(Omega_0)
                    omegas.append(Omega_0)
                    corner_freqs.append(f_c)
                
                if omegas:
                    M_0 = 4.0 * np.pi * density * velocity ** 3 * distance * \
                        np.sqrt(omegas[0] ** 2 + omegas[1] ** 2 + omegas[2] ** 2) / \
                        radiation_pattern
                    r = 3 * k * vs / sum(corner_freqs)
                    moments.append(M_0)
                    source_radii.append(r)
                    corner_frequencies.extend(corner_freqs)
                else:
                    pass
            if not len(moments):
                print ("No moments could be calculated for event") 
                # print ("No moments could be calculated for event %s") % \
                #     event.resource_id.resource_id
                continue

            # Calculate the seismic moment via basic statistics.
            moments = np.array(moments)
            moment = moments.mean()
            moment_std = moments.std()

            corner_frequencies = np.array(corner_frequencies)
            corner_frequency = corner_frequencies.mean()
            corner_frequency_std = corner_frequencies.std()

            # Calculate the source radius.tation_code
            source_radii = np.array(source_radii)
            source_radius = source_radii.mean()
            source_radius_std = source_radii.std()

            # Calculate the stress drop of the event based on the average moment and
            # source radii.
            stress_drop = (7 * moment) / (16 * source_radius ** 3)
            stress_drop_std = np.sqrt((stress_drop ** 2) * \
                (((moment_std ** 2) / (moment ** 2)) + \
                (9 * source_radius * source_radius_std ** 2)))
            # if source_radius > 0 and source_radius_std < source_radius:
            #     print ("Source radius:", source_radius, " Std:", source_radius_std)
            #     print ("Stress drop:", stress_drop / 1E5, " Std:", stress_drop_std / 1E5)

            Mw = 2.0 / 3.0 * (np.log10(moment) - 9.1)
            Mw_std = 2.0 / 3.0 * moment_std / (moment * np.log(10))
            Mws_std.append(Mw_std)
            Mws.append(Mw)
            # Mls.append(local_magnitude)
            # calc_diff = abs(Mw - local_magnitude)
            # Mw = ("%.3f" % Mw).rjust(7)
            # Ml = ("%.3f" % local_magnitude).rjust(7)
            # diff = ("%.3e" % calc_diff).rjust(7)
            # ret_string = colorama.Fore.GREEN + \
            #     "For event %s: Ml=%s | Mw=%s | " % (event.resource_id.resource_id,
            #     Ml, Mw)
            # if calc_diff >= 1.0:
                # ret_string += colorama.Fore.RED
            # ret_string += "Diff=%s" % diff
            # ret_string += colorama.Fore.GREEN
            # ret_string += " | Determined at %i stations" % len(moments)
            # ret_string += colorama.Style.RESET_ALL
            # print (ret_string)
            # print("Mw", Mw)
            print(f"{origin_time}: ({latitude},{longitude},{depth}) | Mw:{round(Mw,3)}")
            mag = Mag()
            mag.mag = Mw
            mag.mag_errors.uncertainty = Mw_std
            mag.magnitude_type = "Mw"
            mag.origin_id = event.origins[0].resource_id
            mag.method_id = "smi:com.github/krischer/moment_magnitude_calculator/automatic/1"
            mag.station_count = len(moments)
            mag.evaluation_mode = "automatic"
            mag.evaluation_status = "preliminary"
            mag.comments.append(Comment( \
                "Seismic Moment=%e Nm; standard deviation=%e" % (moment,
                moment_std)))
            mag.comments.append(Comment("Custom fit to Boatwright spectrum"))
            if source_radius > 0 and source_radius_std < source_radius:
                mag.comments.append(Comment( \
                    "Source radius=%.2fm; standard deviation=%.2f" % (source_radius,
                    source_radius_std)))
            
            event.magnitudes.append(mag)
            event.preferred_magnitude_id = mag.resource_id

        if outfile != None:
            print ("Writing output file...")
            cat.write(outfile, out_format)
        return cat

    def get_corresponding_stream(self,waveform_id, pick_time, padding=1.0,
                                waterlevel=10):
        """
        Helper function to find a requested waveform in the previously created
        waveform_index file.
        Also performs the instrument correction.
        Returns None if the file could not be found.
        """
        
        # from obspy.clients.filesystem.sds import Client as SDS_Client
        # sds_archive = "/home/emmanuel/archive/sds"
        # client = SDS_Client(sds_archive,
        #             sds_type='D', format='MSEED')
        start = pick_time - padding
        end = pick_time + padding
        try:
            st= self.client.get_waveforms(waveform_id.network_code,
                                    waveform_id.station_code,
                                    "*",
                                    waveform_id.channel_code[0:2]+"*",start,end
                                    )
            print(f"OK: {waveform_id.network_code}-{waveform_id.station_code}"+\
                    f"-*-{waveform_id.channel_code[0:2]}* |"+\
                     f"{start} - {end}  ")
        except:
            print(f"Not found: {waveform_id.network_code}-{waveform_id.station_code}"+\
                    f"-*-{waveform_id.channel_code[0:2]}* |"+\
                     f"{start} - {end}  ")
            return None

        for trace in st:
            parsers = self._get_channel_resp()
            # print(trace,parsers)
            paz = parsers[trace.id].get_paz(trace.id, start)
            print("paz",paz)
            # PAZ in SEED correct to m/s. Add a zero to correct to m.
            paz["zeros"].append(0 + 0j)
            trace.detrend()
            trace.simulate(paz_remove=paz, water_level=waterlevel)
        return st

    def estimate_local_magnitude(self,outfile=None,padding=0.5,
                        out_format="QUAKEML",n_processor=None):
        cat = self.catalog
        
        tic = time.time()

        
        # def local_main(event):
        for event in cat:
            if not event.origins:
                print ("No origin for event %s" % event.resource_id)
                continue
            # if not event.magnitudes:
            #     print ("No magnitude for event %s" % event.resource_id)
            #     continue

            origin = event.preferred_origin()
            lat = origin.latitude
            lon = origin.longitude
            depth = origin.depth
            if lat ==None or lon==None:
                continue
            # local_magnitude = event.magnitudes[0].mag

            #if local_magnitude < 1.0:
                #continue
            times2trim={}
            for pick in event.picks:
                if pick.phase_hint.lower() == "s":
                    wav_id = pick.waveform_id

                    if self.jsonfile != None:
                        net = self.info_from_json[wav_id.station_code]["network"]
                        cha = self.info_from_json[wav_id.station_code]["channels"][0][:2]+"*"
                    else: 
                        net = wav_id.network_code
                        cha = wav_id.channel_code[:2]+"*"


                    key = ".".join((net,
                                    wav_id.station_code,"*",
                                    cha))
                    start = pick.time - padding
                    end = pick.time + padding
                    times2trim[key]=(start,end)
            if not times2trim:
                continue

            Ml = self._estimate_local_CC_magnitude((lat,lon,depth),times2trim,origin.time)
                # print(key)

            mag = Mag()
            mag.mag = Ml
        #     mag.mag_errors.uncertainty = Mw_std
            mag.magnitude_type = "Ml"
            mag.origin_id = event.origins[0].resource_id
            mag.method_id = "smi:com.github/local_mag"
            mag.station_count = len(times2trim)
            mag.evaluation_mode = "automatic"
            mag.evaluation_status = "preliminary"
            mag.comments.append(Comment( \
                "local Moment=%e Nm" % (Ml)))
            event.magnitudes.append(mag)
            event.preferred_magnitude_id = mag.resource_id

        # with cf.ThreadPoolExecutor(n_processor) as executor:
        #     executor.map(local_main,cat)

        toc = time.time()
        print(toc-tic,"exe")
        if outfile != None:
            print ("Writing output file...")
            cat.write(outfile, out_format)
        return cat

    def estimate_moment_magnitude(self,output_file=None,density=2650.0,
                        vp=4266.96,
                        vsp_factor=1.84,time_before_pick=0.2,
                        time_after_pick=0.8,padding=20,
                        waterlevel=10.0,out_format="QUAKEML"):

        cat = self._estimate_moment_magnitudes(output_file,
                        vp,vsp_factor,density,time_before_pick,
                        time_after_pick,padding,
                        waterlevel,out_format)
        return cat


if __name__ =="__main__":
    from obspy.clients.fdsn import Client
    from obspy.core.inventory.inventory import read_inventory
    client = 'http://sismo.sgc.gov.co:8080'
    client = Client(client)

    resp = "/home/emmanuel/Ecopetrol/SeisMonitor/data/metadata/CM.dataless"
    p = Parser(resp)
    print(p)
    seed_id = "CM.TAPM.00.HHN"
    psi = p.get_paz(seed_id)
    print(psi)

    # resp = "/home/emmanuel/EDCT/SeisMonitor/data/events/public_CM.xml"
    # p = read_inventory(resp)
    # seed_id = "CM.BAR2.00.HHN"
    # psi = p.get_response(seed_id,"2022-05-01T00:00:00")
    # paz = psi.get_paz()
    # sensitivity = psi.instrument_sensitivity.value
    # gain = paz.normalization_factor
    # poles = paz.poles
    # zeros = paz.zeros

    # paz = {'poles': poles,
    #         'zeros': zeros,
    #         'gain': gain,
    #         'sensitivity': sensitivity}

    # st = client.get_waveforms("CM","BAR2","00","HHN",
    #                         UTCDateTime(2022,5,1,0,0,0),
    #                         UTCDateTime(2022,5,1,0,3,0))
    # tr = st[0]
    # print(tr.data)
    # x = tr.simulate(paz_remove=paz, water_level=10)
    # print(x.data)

    # catalog = "/home/emmanuel/EDCT/SeisMonitor/data/events/SGC2022knomqj.xml"
    # resp = "/home/emmanuel/Ecopetrol/SeisMonitor/data/metadata/CM.dataless"
    # # resp = "/home/emmanuel/EDCT/SeisMonitor/data/events/public_CM.xml"
    # out="./test_magnitude,xml"
    # mag = Magnitude(client,catalog,resp,
    #             jsonfile=None) 
    # mag.estimate_moment_magnitude(out)
    # print(mag)



import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import os
import glob
import shutil
from SeisMonitor.monitor.downloader.utils import get_chunktimes
from SeisMonitor.monitor.downloader.seismonitor import MseedDownloader
from SeisMonitor.monitor.downloader.utils import sanitize_provider_times
from SeisMonitor.monitor.picker import ai as ai_picker
from SeisMonitor.monitor.associator import ai as ai_asso
from SeisMonitor.monitor.locator.nlloc import nlloc
from SeisMonitor.monitor.magnitude.mag import Magnitude

def get_preproc_providers(providers,chunklength_in_sec,
                            out_folder):
    sanitize_provider_times(providers)
    oneprovider = providers[0]
    restrictions = oneprovider.waveform_restrictions
    starttime = restrictions.starttime
    endtime = restrictions.endtime
    chunktimes = get_chunktimes(starttime,endtime,
                                chunklength_in_sec)
    preproc_providers = []
    for starttime,endtime in chunktimes:
        folders = get_folders_by_chunk(out_folder,
                                        starttime,endtime)
        # print("folders")

        new_providers = []
        for provider in providers:
            # provider = provider.copy()
            provider.waveform_restrictions.starttime = starttime
            provider.waveform_restrictions.endtime = endtime
            new_providers.append(provider)

        # providers = new_providers

        chunk_provider = {"providers":new_providers,
                        "folders":folders,
                        "dates":{"starttime":starttime,"endtime":endtime}}
        preproc_providers.append(chunk_provider)

   
    # import time
    # print("sleep")
    # time.sleep(60)
    # exit()
    return preproc_providers



def get_chunktimes_by_provider(providers,chunklength_in_sec):
    times_by_provider = {}
    for provider in providers:
        restrictions = provider.waverform_restrictions
        starttime = restrictions.starttime
        endtime = restrictions.endtime

        times = get_chunktimes(starttime,endtime,
                                chunklength_in_sec)
        times_by_provider.append(times)
    return times_by_provider

def get_folders_by_chunk(out_folder,starttime,endtime):
    st = starttime.strftime("%Y%m%dT%H%M%S")
    et = endtime.strftime("%Y%m%dT%H%M%S")
    chunk_name = st+"__"+et
    chunk_dir = os.path.join(out_folder,chunk_name)

    downloads_dir = os.path.join(chunk_dir,'downloads')
    metadata_dir = os.path.join(chunk_dir,'metadata')
    detections_dir = os.path.join(chunk_dir,'detections')
    asso_dir = os.path.join(chunk_dir,'associations')
    loc_dir = os.path.join(chunk_dir,'locations')
    events_dir = os.path.join(chunk_dir,'magnitudes')

    return {"metadata":metadata_dir,
            "downloads":downloads_dir,
            "detections":detections_dir,
            "associations":asso_dir,
            "locations":loc_dir,
            "magnitudes":events_dir}

def sanitize_pick_batch_size(pickers,download_args):
    overlaps = []
    batch_sizes = []
    for picker,args in pickers.items():
        batch_sizes.append(args.batch_size)
        overlaps.append(args.overlap)

    download_args["picker_args"]["overlap"] = min(overlaps)
    download_args["picker_args"]["batch_size"] = min(batch_sizes)
    return download_args

def sanitize_downloads(pickers):
    new_pickers = {}
    for i,(picker,args) in enumerate(pickers.items()):
        if i == (len(pickers.keys())-1):
            pass
        else:
            args.rm_download = False
        new_pickers[picker] = args
    return new_pickers


class SeisMonitor():
    def __init__(self,providers,out_folder,
                chunklength_in_sec=3600,
                overwrite=False):
        self.providers = providers
        self.out_folder = out_folder
        self.chunklength_in_sec = chunklength_in_sec

        # self.download_folder = os.path.join("%s","downloads")
        # self.metadata_folder = os.path.join("%s","metadata")
        # self.pick_folder = os.path.join("%s","detections")
        # self.association_folder = os.path.join("%s","associations")
        # self.location_folder = os.path.join("%s","locations")
        # self.mag_folder = os.path.join("%s","magnitudes")
        self.overwrite = overwrite
        self.process = {}

    def add_downloader(self,
                    threshold= 60,
                    overlap_in_sec=0,
                    picker_args= {},
                    groupby='{network}.{station}.{channel}',
                    n_processor=None):
        dld_args = locals().copy()
        dld_args["chunklength_in_sec"] = self.chunklength_in_sec
        dld_args.pop("self")
        self.process["downloader"] = dld_args

        
    def add_picker(self,
                    pickers={}):

        if pickers:
            pickers = sanitize_downloads(pickers)
            self.process["picker"] = pickers
            if "downloader" in list(self.process.keys()):
                self.process["downloader"] = sanitize_pick_batch_size(pickers,self.process["downloader"])
        
        self.picker_output = list(pickers.keys())
        return list(pickers.keys())

    def add_associator(self,input,
                        associators={}):
        self.associator_input = input
        if associators:
            self.process["associator"] = associators

        out = {}
        for associator in associators.keys():
            for picker in self.associator_input:
                name = "_".join((associator,picker))
                out[name] = os.path.join(associator,picker)
        self.associator_output = out
        return out

    def add_locator(self,
                    input,
                    locators={}
                    ):
        self.locator_input = input

        if locators:
            self.process["locator"] = locators

        out = {}
        for locator in locators.keys():
            for task,project in self.locator_input.items():
                assert task in ["associations","locations","magnitudes"]
                assert isinstance(project,tuple)
                assert len(project)==2

                out_name = "_".join((locator,task,project[0],project[1]))
                out[out_name] = os.path.join(locator,task,project[0],project[1])

        self.locator_output = out
        return out

    def add_magnitude(self,
                    input,
                    magnitudes={}
                    ):
        self.magnitude_input = input

        if magnitudes:
            self.process["magnitude"] = magnitudes

        out = {}
        for magnitude in magnitudes.keys():
            for task,project in self.magnitude_input.items():
                assert task in ["associations","locations","magnitudes"]
                assert isinstance(project,tuple)
                assert len(project)==2

                out_name = "_".join((magnitude,task,project[0],project[1]))
                out[out_name] = os.path.join(magnitude,task,project[0],project[1])

        self.magnitude_output = out
        return out

    def run(self):
        preproc_providers = get_preproc_providers(self.providers,
                                            self.chunklength_in_sec,
                                            self.out_folder)
        # print("preproc_providers")

        for chunk_provider in preproc_providers:
            print("chunk:",chunk_provider["dates"]["starttime"],
                "--",chunk_provider["dates"]["endtime"])
            # print(chunk_provider)
            providers = chunk_provider["providers"]
            folders = chunk_provider["folders"]

            # for provider in providers:
            #     wav = provider.waveform_restrictions
            #     print(provider.__dict__)
            
            # exit()

            for process, process_args in self.process.items():

                if process == "downloader":
                    structure = os.path.join("{station}","{network}.{station}.{location}.{channel}__{starttime}__{endtime}.mseed")
                    download_path = os.path.join(folders["downloads"],structure)
                    md = MseedDownloader(providers)
                    md.make_inv_and_json(folders["metadata"])
                    md.download(download_path,**process_args)
                    del md

                elif process == "picker":
                    for picker,picker_args in process_args.items():
                        out_path = os.path.join(folders["detections"],picker)
                        if picker == "EQTransformer":
                            _picker = ai_picker.EQTransformer(picker_args)
                            result = _picker.pick(folders["downloads"],
                                                folders["metadata"],
                                                out_path)
                            if result.empty:
                                print("No picks")
                                continue
                            del _picker
                            del result

                        elif picker == "PhaseNet":
                            _picker = ai_picker.PhaseNet(picker_args)
                            result = _picker.pick(folders["downloads"],
                                                folders["metadata"],
                                                out_path)
                            if result.empty:
                                print("No picks")
                                continue
                            del _picker
                            del result
                            
                elif process == "associator":
                    inv = os.path.join(folders["metadata"],"inv.xml")

                    for picker in self.associator_input:
                        picker_path = os.path.join(folders["detections"],picker)
                        picks_path = os.path.join(picker_path,"results",
                                                "seismonitor_picks.csv")
                        for associator,associator_args in process_args.items():
                            out_name = self.associator_output[f"{associator_args.name}_{picker}"]
                            out_folder = os.path.join(folders["associations"],out_name)
                            if associator == "GaMMA":
                                _associator = ai_asso.GaMMA(associator_args)
                                _,result,_ = _associator.associate(picks_path,
                                                            inv,out_folder)
                                if result.empty:
                                    print("No associated picks")
                                    continue

                elif process == "locator":
                    for locator,locator_args in process_args.items():
                        for task,project in self.locator_input.items():
                            catalog = os.path.join(folders[task],project[0],project[1],task+".xml")
                            nlloc_folder = os.path.join(folders["locations"],locator,project[0],project[1])
                            locator_args.locate(catalog,nlloc_folder)

                elif process == "magnitude":
                    for magnitude,magnitude_args in process_args.items():
                        for task,project in self.magnitude_input.items():
                            catalog = os.path.join(folders[task],project[0],project[1],task+".xml")
                            mag_folder = os.path.join(folders["magnitudes"],magnitude,project[0],project[1])
                            mag = Magnitude(providers=self.providers,catalog=catalog,
                                            out_dir=mag_folder)
                            if magnitude == "Ml":
                                mag.get_Ml(**magnitude_args)











                
        
        
        




        # self.json_path = os.path.join(self.download_folder,"json",
        #                                             "stations.json")
        # self.mseed_folder = os.path.join(self.download_folder,"mseed")
        # if object.name == "EQTransformer":
        #     picker = EQTransformer(self.mseed_storage,
        #                            self.json_path,out_dir)

# if __name__ == "__main__":



from genericpath import isdir
from . import utils as ut
import shutil
import os
import time
import logging
import warnings
import pandas as pd
import datetime as dt
import tensorflow as tf
from SeisMonitor.utils import printlog, isfile
from EQTransformer.core.mseed_predictor import mseed_predictor

warnings.simplefilter(action='ignore', category=FutureWarning)
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

# EQTransformer_core_path = os.path.join(os.path.dirname(__file__),"core","EQTransformer")
# EQTransformer_model_path = os.path.join(EQTransformer_core_path,"ModelsAndSampleData","EqT_model.h5")
# PhaseNet_core_path = os.path.join(os.path.dirname(__file__),"core","PhaseNet")
# PhaseNet_model_path = os.path.join(PhaseNet_core_path,"model","190703-214543")

class EQTransformerObj(object):
    def __init__(self,
                # eqt_path,
                model_path,
                n_processor=2,overlap=0.3,
                detection_threshold=0.1, P_threshold=0.1,
                S_threshold=0.1,number_of_plots=1,
                batch_size=1,
                plot_mode=1,
                overwrite=False,
                # chunk_size=3600,
                rm_downloads=False):

        """
        EQTransformer parameters
        """
        # self.eqt_path = eqt_path
        # self.model_path = os.path.join(eqt_path,"ModelsAndSampleData","EqT_model.h5")
        self.model_path = model_path
        self.n_processor = n_processor
        self.overlap = overlap
        self.detection_threshold = detection_threshold
        self.P_threshold = P_threshold
        self.S_threshold = S_threshold
        self.number_of_plots = number_of_plots
        self.batch_size = batch_size
        self.plot_mode = plot_mode
        self.overwrite = overwrite
        self.rm_downloads = rm_downloads
        self.name = "EQTransformer"

class PhaseNetObj(object):
    def __init__(self, pnet_path,
                 mode='pred', P_threshold=0.3, S_threshold=0.3,
                batch_size=2, plot = False, save_result=False,
                epochs = 100,learning_rate= 0.01,decay_step = -1,
                decay_rate = 0.9,momentum = 0.9,filters_root = 8,
                depth = 5, kernel_size = [7, 1], pool_size = [4, 1],
                drop_rate = 0, dilation_rate = [1, 1],loss_type = "cross_entropy",
                weight_decay = 0, optimizer = "adam",   summary = True,
                class_weights = [1, 1, 1],log_dir = "log",  num_plots = 10,
                input_length = None,input_mseed = True, filename_picks = "picks",
                one_single_sampling_rate=-1,
                data_dir = "./dataset/waveform_pred/",
                data_list = "./dataset/waveform.csv",
                train_dir = "./dataset/waveform_train/",
                train_list ="./dataset/waveform.csv",
                valid_dir = None, valid_list = None,
                output_dir = None,
                rm_downloads=False ):
        """
        PhaseNet parameters
        """

        self.phasenet_path = pnet_path
        self.model_dir = os.path.join(pnet_path,"model","190703-214543")
        self.mode = mode
        self.tp_prob = P_threshold
        self.ts_prob = S_threshold
        self.batch_size = batch_size
        self.plot_figure = plot
        self.save_result = save_result
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.decay_step = decay_step
        self.decay_rate = decay_rate
        self.momentum = momentum
        self.filters_root = filters_root
        self.depth = depth
        self.kernel_size = kernel_size
        self.pool_size = pool_size
        self.drop_rate = drop_rate
        self.dilation_rate = dilation_rate
        self.loss_type = loss_type
        self.weight_decay = weight_decay
        self.optimizer = optimizer
        self.summary = summary
        self.class_weights = class_weights
        self.log_dir = log_dir
        self.num_plots = num_plots
        self.input_length = input_length
        self.input_mseed = input_mseed
        self.fpred = filename_picks
        self.data_dir = data_dir
        self.data_list = data_list
        self.train_dir = train_dir
        self.train_list = train_list
        self.valid_dir = valid_dir
        self.valid_list = valid_list
        self.output_dir = output_dir
        self.overlap = 0.5
        self.rm_downloads = rm_downloads
        self.one_single_sampling_rate = one_single_sampling_rate
        self.name = "PhaseNet"

class EQTransformer():
    def __init__(self,eqt_obj):
        self.eqt_obj = eqt_obj
    
    def pick(self,mseed_storage,metadata_dir,
                out_dir):
        """
        Parameters:
        --------------
        eqt_obj: ./utils/eqt.EQTobj
            Object where you can set your own
            EQTransformer parameter
        rm_downloads: bool
            True for remove the input mseed folder

        Returns:
        ---------
        It saves your EQtransformer solutions in the next path:
        {self.picks_dir}/eqt
        """
        self.mseed_storage = mseed_storage
        self.json_path = os.path.join(metadata_dir,"stations.json")
        self.pick_storage = os.path.join(out_dir,"results")
        self.msg_author = "Picker: EQTransfomer"

        tic = time.time()
        printlog('info',self.msg_author,'Running EQTransformer')
        try:
            tf.compat.v1.reset_default_graph()
            mseed_predictor(
                    input_dir=self.mseed_storage,
                    input_model=self.eqt_obj.model_path,
                    stations_json=self.json_path,
                    output_dir=self.pick_storage,
                    detection_threshold=self.eqt_obj.detection_threshold,
                    P_threshold=self.eqt_obj.P_threshold, 
                    S_threshold=self.eqt_obj.S_threshold,
                    number_of_plots=self.eqt_obj.number_of_plots, 
                    plot_mode=self.eqt_obj.plot_mode, 
                    overlap=self.eqt_obj.overlap, 
                    batch_size=self.eqt_obj.batch_size,
                    overwrite = self.eqt_obj.overwrite)
        except ValueError as e:
            if str(e) == ("The target structure is of type `<class 'NoneType'>`\n"+
                        "  None\n"+
                        "However the input structure is a sequence (<class 'list'>) of length 0.\n"+
                        "  []\n"+
                        "nest cannot guarantee that it is safe to map one to the other."):
                printlog('info',self.msg_author,
                f"Error in batch_size: {self.eqt_obj.batch_size} to the"+
                " last mseed. Check 'to_pick' parameter in DownloadRestrictions")
        # eqt_merge = os.path.join(self.pick_storage,"eqt_merge.csv")
        # ut.merge_picks(self.pick_storage,"X_prediction_results.csv",
                    #    eqt_merge)
        output_path = os.path.join(self.pick_storage,"seismonitor_picks.csv")
        result = ut.eqt_picks_2_seismonitor_fmt(self.pick_storage,self.mseed_storage,
                                        output_path)
        tf.keras.backend.clear_session()   
        toc = time.time()
        exetime = dt.timedelta(seconds=toc-tic)
        printlog('info',self.msg_author,
            f'Total time of execution: {exetime.total_seconds()} seconds')

        if self.eqt_obj.rm_downloads:
            shutil.rmtree(self.mseed_storage, ignore_errors=True)
        return result

class PhaseNet():
    def __init__(self,pnet_obj):
        self.pnet_obj = pnet_obj
        self.msg_author = "Picker: PhaseNet"

    def __mv_downloads2onefolder(self):
        """
        Move all downloads to folder to all_mseed folder.
        """
        tic = time.time()
        # ut.get_one_stream(self.mseed_storage,
        #                 self.datadir)
        if not os.path.isdir(self.datadir):
            os.makedirs(self.datadir)

        ut.mv_mseed2onefolder(self.mseed_storage,self.datadir)
        # ut.mv_mseed2stationfolder(self.datadir,self.mseed_storage)
        # exit()
        toc = time.time()
        exetime = dt.timedelta(seconds=toc-tic)
        printlog("info",self.msg_author,f"Moving mseed2onefolder in {exetime.total_seconds()} seconds.")

    def __mv_downloads2stationfolder(self):
        """
        Move all downloads to folder to all_mseed folder.
        """
        tic = time.time()
        ut.mv_mseed2stationfolder(self.datadir,self.mseed_storage)
        toc = time.time()
        exetime = dt.timedelta(seconds=toc-tic)
        printlog("info",self.msg_author,f"Moving mseed2stationfolder in {exetime.total_seconds()} seconds.")

    def __make_datalist(self):
        tic = time.time()
        printlog("info",self.msg_author,"Running to create datalist")
        
        datalist = ut.make_PhaseNet_datalist(datadir=self.datadir, 
                                json_path=self.json_path,
                                datalist_path=self.datalist)

        toc = time.time()
        exetime = dt.timedelta(seconds=toc-tic)
        printlog("info",self.msg_author,f"Datalist created in {exetime.total_seconds()} seconds.")

    def pick(self,mseed_storage,metadata_dir,
                out_dir):
        self.mseed_storage = mseed_storage
        self.json_path = os.path.join(metadata_dir,"stations.json")
        self.pick_storage = os.path.join(out_dir,"results")
        self.datadir = os.path.join(out_dir,"datadir")
        self.datalist = os.path.join(out_dir,"datalist",'fname.csv')
        
        
        self.__mv_downloads2onefolder()
        self.__make_datalist()

        tf.compat.v1.reset_default_graph()
        tf.keras.backend.clear_session()

        tic = time.time()
        printlog("info",self.msg_author,"Running PhaseNet pretrained model.")

        args = self.pnet_obj
        args.data_dir = self.datadir
        args.data_list = self.datalist
        args.output_dir  = self.pick_storage

        ut.phasenet_from_console(args,self.msg_author)
        datapicks = os.path.join(self.pick_storage,'picks.csv')
        seismonitor_datapicks = os.path.join(self.pick_storage,'seismonitor_picks.csv')
        result = ut.get_picks(datapicks=datapicks,
                            datalist=self.datalist,
                            min_p_prob=self.pnet_obj.tp_prob, 
                            min_s_prob=self.pnet_obj.ts_prob, 
                            one_single_sampling_rate = self.pnet_obj.one_single_sampling_rate,
                             mode='df_obj',
                            export=seismonitor_datapicks)

        try:
            ut.rm_phasenet_duplicate_picks(path=seismonitor_datapicks, 
                                        output_path=seismonitor_datapicks)
            printlog('info',self.msg_author,
                    f'Duplicated picks had been removed. See your results in {seismonitor_datapicks}')
        except Exception as e:
            printlog('error',self.msg_author,
                    f" Can't remove duplicated picks: {e}")
        
        toc = time.time()
        exetime = dt.timedelta(seconds=toc-tic)
        printlog("info",self.msg_author,f'Time of prediction: {exetime.total_seconds()} seconds')

        self.__mv_downloads2stationfolder()        
        if self.pnet_obj.rm_downloads:
            shutil.rmtree(self.mseed_storage, ignore_errors=True)

        return result


from . import utils as ut
from obspy import read_inventory
import copy

class Provider():
	def __init__(self,client,waveform_restrictions, 
				processing=None,xml=None) -> None:
		self.client = client
		self.waveform_restrictions = waveform_restrictions
		self.processing = processing
		self.xml = xml

		if xml != None:
			self.inventory = read_inventory(xml)
		else:
			self.inventory = client.get_stations(network=waveform_restrictions.network,
												station=waveform_restrictions.station,
												location="*",
												channel="*",level='response')


	def copy(self):
		return copy.deepcopy(self)

class WaveformRestrictions():
	def __init__(self,network,station,
			  location,channel,
			  starttime,endtime,
			  location_preferences=[],
			  channel_preferences=[],
			  filter_networks=[], 
			  filter_stations=[],
			  filter_domain=[-180,180,-90,90] #lonw,lone,lats,latn
			  ):
		"""
		network: str
			Select one or more network codes. 
			Multiple codes are comma-separated (e.g. "IU,TA"). 
			Wildcards are allowed.
		station: str
			Select one or more SEED station codes. 
			Multiple codes are comma-separated (e.g. "ANMO,PFO"). 
			Wildcards are allowed.
		location: str
			Select one or more SEED location identifiers. 
			Multiple identifiers are comma-separated (e.g. "00,01"). 
			Wildcards are allowed.
		channel: str
			Select one or more SEED channel codes. 
			Multiple codes are comma-separated (e.g. "BHZ,HHZ").
		starttime: obspy.UTCDateTime
			Limit results to time series samples on or 
			after the specified start time.
		endtime: obspy.UTCDateTime
			Limit results to time series samples on or 
			before the specified end time.
		location_preferences: list
			list of location in the order of the preference. If select the
			location of the first element, then the rest of elements will not be 
			downloaded.
		"""
		self.network = network
		self.station = station
		self.location = location
		self.channel = channel
		self.starttime = starttime
		self.endtime = endtime
		self.location_preferences = location_preferences
		self.channel_preferences = channel_preferences
		self.filter_networks = filter_networks
		self.filter_stations = filter_stations
		self.filter_domain = filter_domain

class Processing(object):
	def __init__(self,
				order=['normalize','merge','detrend',
						'taper',"filter"],
				decimate={"factor":2},
				detrend={"type":"demean"},
				filter={"type":'bandpass', 
						"freqmin" : 1, 
						"freqmax" : 45, 
						"corners":2, 
						"zerophase":True},
				merge={"method":0,"fill_value":'latest'},
				normalize={"global_max":False},
				resample={"sampling_rate":200},
				taper={"max_percentage":0.001, 
						"type":"cosine", 
						"max_length":2} ,
				select_networks=[], 
				select_stations=[],
				filter_networks=[], 
			  	filter_stations=[]):
		"""
		Restrictions to preprocess a stream selected by seed_ids
		
		Parameters:
		-----------
		order: list of str
			Order to preprocess the stream.
			ex: ['merge','detrend','taper','normalized']
		decimate: dict
			Contains the parameters of the decimate Stream method 
		detrend: dict 
			Contains the parameters of the detrend Stream method 
		filter: dict
			Contains the parameters of the filter Stream method 
		merge: dict
			Contains the parameters of the merge Stream method 
		normalize: dict
			Contains the parameters of the normalize Stream method 
		remove_response: dict
			Contains the parameters of the remove_response Stream method 
		resample: dict
			Contains the parameters of the resample Stream method 
		taper: dict
			Contains the parameters of the taper Stream method 

		--------
		"""
		self.order = order
		self.decimate = decimate
		self.detrend = detrend
		self.filter = filter
		self.merge = merge
		self.normalize = normalize
		self.resample = resample
		self.taper = taper
		self.select_networks = select_networks
		self.select_stations = select_stations
		self.filter_networks = filter_networks
		self.filter_stations = filter_stations
	
	def run(self,st):
		return ut.preproc_stream(st,self.order ,
							self.decimate ,
							self.detrend ,
							self.filter ,
							self.merge ,
							self.normalize ,
							self.resample ,
							self.taper ,
							self.select_networks ,
							self.select_stations ,
							self.filter_networks ,
							self.filter_stations)



"""
NLL runner was mainly based on based on

https://github.com/saeedsltm/PyNLLRunner
"""

import sys
import os
import shutil
import glob
seismopath = "/home/emmanuel/EDCT"
seismonitor = os.path.join(seismopath,"SeisMonitor")
sys.path.insert(0,seismonitor)
from obspy import read_inventory
from subprocess import STDOUT, check_call
import SeisMonitor.utils as ut
from subprocess import run
import pandas as pd
CORE_NLLOC = os.path.join(os.path.dirname(__file__),"cor")

class LocObj():
    def __init__(self,
        nstd              = 1    ,    # number of standard deviation for plotting ellipses 1: %68 , 2: %95.
        lat_min           = 35.00,    # minimum latitude
        lat_max           = 37.00,    # maximum latitude
        lon_min           = 51.00,    # minimum longitude    
        lon_max           = 53.00,    # maximum longitude
        dep_max           = 25.00,    # maximum depth
        p_res_max         = 0.5  ,    # maximum p-residual 
        s_res_max         = 0.5  ,    # maximum s-residual
        max_dist          = 50   ,    # maximum distance for residual plot
        herr_max          = 10   ,    # maximum horizontal error
        zerr_max          = 10   ,    # maximum depth error
        rms_max           = 1.0  ,    # maximum rms 
        minds_max         = 20   ,    # maximum mindistance
        dep_hist_bw       = 2    ,    # depth histogram bin width
        her_hist_bw       = 0.25 ,    # herr histogram bin width
        zer_hist_bw       = 0.25 ,    # zerr histogram bin width
        rms_hist_bw       = 0.05 ,    # rms histogram bin width
    ):

        self.nstd              = nstd    # number of standard deviation for plotting ellipses 1: %68 , 2: %95.
        self.lat_min           = lat_min    # minimum latitude
        self.lat_max           = lat_max    # maximum latitude
        self.lon_min           = lon_min    # minimum longitude    
        self.lon_max           = lon_max    # maximum longitude
        self.dep_max           = dep_max    # maximum depth
        self.p_res_max         = p_res_max    # maximum p-residual 
        self.s_res_max         = s_res_max    # maximum s-residual
        self.max_dist          = max_dist    # maximum distance for residual plot
        self.herr_max          = herr_max    # maximum horizontal error
        self.zerr_max          = zerr_max    # maximum depth error
        self.rms_max           = rms_max    # maximum rms 
        self.minds_max         = minds_max    # maximum mindistance
        self.dep_hist_bw       = dep_hist_bw    # depth histogram bin width
        self.her_hist_bw       = her_hist_bw    # herr histogram bin width
        self.zer_hist_bw       = zer_hist_bw    # zerr histogram bin width
        self.rms_hist_bw       = rms_hist_bw    # rms histogram bin width
#  self.nlloc_cf.write('VGGRID  2 %s %s %s %s %s %s SLOW_LEN\n'%(self.vggrid_num_g_n_xy,
#                                                                       self.vggrid_num_g_n_z,
#                                                                       self.vggrid_grid_x,
#                                                                       self.vggrid_grid_y,
#                                                                       self.vggrid_grid_z,
#                                                                       self.vggrid_g_s_xyz))
class NLLocObj():
    def __init__(self,folder_inp,
        trans = "SIMPLE",
        trans_args = [5,-73,0],
        vggrid_num_nodes = [2,583,70],
        vggrid_coord_orig = [-891,-891,-5],
        vggrid_node_spacing =  [3,0,0],
        locsig            = "user"            ,   # signature.
        loccom            = "comment"         ,   # comments.
        locfiles_typ      = "SEISAN"         ,   # obs type: (nlloc_obs, hypo71, hypoellipse, renass_dep, seisan, hypodd_).
        lochypout         =  4              ,   # output type: (1:save_nlloc_all,2:save_nlloc_sum,3:save_hypo71_sum,4:all)  
        locsearch_min_xyz =  [10, 10, 4]        ,   # search type, initial number of cells in x/y/z directions. 
        locsearch_node    =  [0.01, 10000, 5000],   # search type, minimum node size, max number of nodes, number of scatters.
        locsearch_stp     =  [0, 1]            ,   #  1:use stations density or not  0 , 1:stop on min node size or not  0  .
        locgrid_num_nodes  = [2, 1001, 201]      ,   # number of nodes along x/y/z axis.
        locgrid_coord_orig = [-891,-891,-5],
        locgrid_node_spacing   = [1.0, 1.0, 1.0]     ,   # grid spacing along x/y/z axis.
        locmeth_args = [200,4,-1,-1,1.78,6,-1,0],
        locgau            =  [0.2, 0.0]      ,   # gaussian model error parameters (locgau sigma_t (s), corrlen (km)).
        locgau2           =  [0.01, 0.05, 2.0]  ,   # travel-time dependent gaussian model error parameters (%tt,tmax,tmin).
        locphaseid_p      =  ["P","p","PN","PG","Pn","Pg"],   # p-phase identifier mapping.
        locphaseid_s      =  ["S","s","SN","SG","Sn","Sg"],   # s-phase identifier mapping. to ignore s: ==>  $ .
        locphstat_args = [9999,-1,9999,1,1,9999,-9999,9999],

    ) -> None:
        self.folder_inp = folder_inp
        self.vggrid_vel_inp    = os.path.join(folder_inp,"model.dat")        # velocity model file.
        self.vggrid_sta_inp    = os.path.join(folder_inp,"station.dat")          # station file.
        self.locfiles_obs      = os.path.join(folder_inp,"select.out")        # observation file name  could be nordic. nlloc, ... .
        
        self.trans = trans
        self.trans_args = " ".join(list(map(lambda x: str(x),trans_args)))
        
        self.vggrid_num_nodes = " ".join(list(map(lambda x: str(x),vggrid_num_nodes)))
        self.vggrid_coord_orig = " ".join(list(map(lambda x: str(x),vggrid_coord_orig)))
        self.vggrid_node_spacing =  " ".join(list(map(lambda x: str(x),vggrid_node_spacing)))
        
        self.locsig            = locsig                # signature.
        self.loccom            = loccom                # comments.
        self.locfiles_typ      = locfiles_typ          # obs type: (nlloc_obs, hypo71, hypoellipse, renass_dep, seisan, hypodd_).
        self.lochypout         = lochypout             # output type: (1:save_nlloc_all,2:save_nlloc_sum,3:save_hypo71_sum,4:all)  
        self.locsearch_min_xyz = " ".join(list(map(lambda x: str(x),locsearch_min_xyz)))     # search type, initial number of cells in x/y/z directions. 
        self.locsearch_node    = " ".join(list(map(lambda x: str(x),locsearch_node)))        # search type, minimum node size, max number of nodes, number of scatters.
        self.locsearch_stp     = " ".join(list(map(lambda x: str(x),locsearch_stp)))         #  1:use stations density or not  0 , 1:stop on min node size or not  0  .
        self.locgrid_num_nodes  = " ".join(list(map(lambda x: str(x),locgrid_num_nodes)))      # number of nodes along x/y/z axis.
        self.locgrid_coord_orig = " ".join(list(map(lambda x: str(x),locgrid_coord_orig)))
        self.locgrid_node_spacing   = " ".join(list(map(lambda x: str(x),locgrid_node_spacing)))       # grid spacing along x/y/z axis.
        self.locmeth_args = " ".join(list(map(lambda x: str(x),locmeth_args))) 
        self.locgau            = " ".join(list(map(lambda x: str(x),locgau)))                # gaussian model error parameters (locgau sigma_t (s), corrlen (km)).
        self.locgau2           = " ".join(list(map(lambda x: str(x),locgau2)))               # travel-time dependent gaussian model error parameters (%tt,tmax,tmin).
        self.locphaseid_p      = " ".join(list(map(lambda x: str(x),locphaseid_p)))          # p-phase identifier mapping.
        self.locphaseid_s      = " ".join(list(map(lambda x: str(x),locphaseid_s)))          # s-phase identifier mapping. to ignore s: ==>  $ .
        self.locphstat_args = locphstat_args     # max hypocenter rms to include in ave residual.

        self.NLLOC_path = os.path.join(CORE_NLLOC,"NonLinLoc-main")
        self.src_path = os.path.join(self.NLLOC_path,"src")
        self.bin_path = os.path.join(self.src_path,"bin")
        self.nll_exe_path = os.path.join(self.bin_path,"NLLoc")
        self.loc_folder = os.path.join(self.folder_inp,"loc","SeisMonitor")
        self.model_folder = os.path.join(self.folder_inp,"model","layer")
        self.time_folder = os.path.join(self.folder_inp,"time","layer")

    def write_nlloc_cf(self, out='./nlloc.cf',
                        P_flag=True, 
                        S_flag=False,
                        rm_outs=False):

        """
        taken from https://github.com/saeedsltm/PyNLLRunner
        """

        self.nlloc_cf = open(out, 'w')

        self.nlloc_cf.write('#__________________START GENERIC CONTROL STATEMENTS\n\n')
        self.nlloc_cf.write('CONTROL 1 54321\n')
        self.nlloc_cf.write('TRANS  %s %s \n\n'%(self.trans,
                                            self.trans_args))
        self.nlloc_cf.write('#__________________END\n')
        self.nlloc_cf.write('#__________________START VEL2GRID STATEMENTS\n\n')
        self.nlloc_cf.write(f'VGOUT  {self.model_folder}\n')
        
        if P_flag: 
            self.nlloc_cf.write('VGTYPE P\n')
        if S_flag: 
            self.nlloc_cf.write('VGTYPE S\n')

        self.nlloc_cf.write('VGGRID  %s %s %s SLOW_LEN\n'%(self.vggrid_num_nodes,
                                                                self.vggrid_coord_orig,
                                                                    self.vggrid_node_spacing))
        self.nlloc_cf.write('INCLUDE %s\n\n'%(self.vggrid_vel_inp))
        self.nlloc_cf.write('#__________________END\n')
        self.nlloc_cf.write('#__________________START GRID2TIME STATEMENTS\n\n')

        if P_flag:
            self.nlloc_cf.write(f'GTFILES  {self.model_folder} {self.time_folder} P\n')
        
        if S_flag:
            self.nlloc_cf.write(f'GTFILES  {self.model_folder} {self.time_folder} S\n')

        self.nlloc_cf.write('GTMODE GRID2D ANGLES_YES\n')
        self.nlloc_cf.write('INCLUDE %s\n'%(self.vggrid_sta_inp))
        self.nlloc_cf.write('GT_PLFD  1.0e-3  0\n\n')
        self.nlloc_cf.write('#__________________END\n')


        self.nlloc_cf.write('#__________________START NLDIFFLOC STATEMENTS\n\n')
        self.nlloc_cf.write('LOCSIG %s \n'%(self.locsig))
        self.nlloc_cf.write('LOCCOM %s \n'%(self.loccom))

        self.nlloc_cf.write('LOCFILES %s %s %s %s\n'%(self.locfiles_obs,
                                                    self.locfiles_typ,
                                                    self.time_folder,
                                                    self.loc_folder))
            
        output = ['SAVE_NLLOC_ALL','SAVE_NLLOC_SUM','SAVE_HYPO71_SUM']
        output.append(' '.join(output))

        self.nlloc_cf.write('LOCHYPOUT %s\n'%(output[int(self.lochypout)-1]))
               
        self.nlloc_cf.write('LOCSEARCH OCT %s %s %s \n'%(self.locsearch,
                                                         self.locsearch_node,
                                                         self.locsearch_stp))

        self.nlloc_cf.write('LOCGRID %s %s %s %s %s PROB_DENSITY SAVE\n'%(self.locgrid_gnum_xyz ,
                                                                          self.locgrid_grid_x,
                                                                          self.locgrid_grid_y,
                                                                          self.locgrid_grid_z,
                                                                          self.locgrid_g_s_xyz))
       
        self.nlloc_cf.write('LOCMETH EDT_OT_WT %s %s %s %s %s %s %s %s\n'%(self.locmeth_max_st_d,
                                                                            self.locmeth_min_nm_ph,
                                                                            self.locmeth_max_nm_ph,
                                                                            self.locmeth_min_nm_s,
                                                                            self.locmeth_vp_vs,
                                                                            self.locmeth_max_g,
                                                                            self.locmeth_min_st_d,
                                                                            self.locmeth_dup))



        self.nlloc_cf.write('LOCGAU %s\n'%(self.locgau))
        self.nlloc_cf.write('LLOCGAU2 %s\n'%(self.locgau2))
        self.nlloc_cf.write('LOCPHASEID P %s\n'%(self.locphaseid_p))
        self.nlloc_cf.write('LOCPHASEID S %s\n'%(self.locphaseid_s))
        self.nlloc_cf.write('LOCQUAL2ERR 0.1 0.5 1.0 2.0 99999.9\n')
        self.nlloc_cf.write('LOCPHSTAT %s %s %s %s %s %s %s %s\n'%(self.locphstat_rms_max,
                                                                   self.locphstat_nr_min,
                                                                   self.locphstat_gap_max,
                                                                   self.locphstat_p_rmax,
                                                                   self.locphstat_s_rmax,
                                                                   self.locphstat_el3_max,
                                                                   self.locphstat_d_min,
                                                                   self.locphstat_d_max))


        self.nlloc_cf.write('LOCANGLES ANGLES_YES 5\n')
        self.nlloc_cf.write('LOCMAG ML_HB 1.0 1.110 0.00189\n')

        self.nlloc_cf.write('\n#__________________END')
            
        self.nlloc_cf.close()

        for _folder in [self.loc_folder,self.model_folder,self.time_folder]:
            folder = os.path.dirname(_folder)
            if os.path.isdir(folder):
                if rm_outs:
                    shutil.rmtree(folder)
                    os.makedirs(folder)
                else:
                    pass
            else:
                os.makedirs(folder)


def apt_install(pkgs):
    cmd = ['pkexec', 'apt-get', 'install', '-y'] + pkgs
    print('Running command: {}'.format(' '.join(cmd)))
    result = run(
        cmd,
        stdout=sys.stdout,
        stderr=sys.stderr,
        encoding='utf8',
        env={**os.environ, 'DEBIAN_FRONTEND': 'noninteractive'}
    )
    result.check_returncode()

def download_nlloc(forced=False):
    name = "nll.zip"
    zip_path = os.path.join(CORE_NLLOC,name)
    NLLOC_path = os.path.join(CORE_NLLOC,"NonLinLoc-main")
    src_path = os.path.join(NLLOC_path,"src")
    bin_path = os.path.join(src_path,"bin")
    nll_exe_path = os.path.join(bin_path,"NLLoc")
    cache_path = os.path.join(src_path,"CMakeCache.txt")
    # "https://github.com/alomax/NonLinLoc/archive/refs/heads/main.zip"

    if not forced:
        if os.path.isfile(nll_exe_path):
            return True
        else:
            pass

    isfile = ut.isfile(zip_path)
    if not isfile:
        os.system(f"wget https://github.com/alomax/NonLinLoc/archive/refs/heads/main.zip -O {zip_path}")

    if not os.path.isdir(NLLOC_path):
        os.system(f"unzip {zip_path} -d {CORE_NLLOC}")

    if os.path.isdir(bin_path):
        os.rmdir(bin_path)
        os.makedirs(bin_path)

    if os.path.isdir(cache_path):
        os.rmdir(cache_path)

    apt_install(["cmake"])

    os.system(f"cd {src_path} && cmake . && make")

def write_1d_vel_model(vel_path,out,
                        compute_vs=True,
                        vp_vs_ratio=1.78):
    df = pd.read_csv(vel_path)

    vm = open(out, 'w')
    msg ="# model layers (LAYER depth, Vp_top, Vp_grad, Vs_top, Vs_grad, p_top, p_grad)\n"
    vm.write(msg)
    for i,row in df.iterrows():
        if compute_vs:
            vs = row.vp/vp_vs_ratio
        else:
            vs = row.vs

        if i == len(df)-1:
            enter = ""
        else:
            enter = "\n"

        msg = f"LAYER    {row.depth:<6.2f}    {row.vp:<.2f}    0.00    {vs:<.2f}    0.00    {row.rho:<.2f}    0.00{enter}"
        vm.write(msg)
    vm.close()

def resp2df(resp):
    """
    Parameters:
    -----------
    resp: str
        RESP filepath

    Returns: DataFrame
        Dataframe with the next columns
        network,station,latitude,longitude,elevation
    """
    networks = []
    stations = []
    longitudes = []
    latitudes = []
    elevations = []
    inv = read_inventory(resp)
    for net in inv:
        for sta in net:
            latitudes.append(sta.latitude)
            longitudes.append(sta.longitude)
            elevations.append(sta.elevation)
            stations.append(sta.code)
            networks.append(net.code)

    df = {"network":networks,"station":stations,
        "latitude":latitudes,"longitude":longitudes,
        "elevation":elevations}
    df = pd.DataFrame(df)
    # print(df)
    return df 

def write_station_file(sta_path,out):
    df = resp2df(sta_path)
    print(df)

    vs = open(out, 'w')
    msg = "# GTSRCE label LATLON latSrce longSrce zSrce elev\n"
    vs.write(msg)
    for i,row in df.iterrows():
        elv = row.elevation/1e3
        if i == len(df)-1:
            enter = ""
        else:
            enter = "\n"

        msg = f"GTSRCE  {row.station:<5}  LATLON  {row.latitude:<7.3f}  {row.longitude:<8.3f}  0.000  {elv:<7.3f}{enter}"
        vs.write(msg)
    vs.close()



if __name__ == "__main__":
    # from obspy.core.event import read_events
    # cat_path = "/home/emmanuel/EDCT/test/associations/associations.xml"
    # out="/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test/select.out"
    # cat = read_events(cat_path)
    # cat.write(out,format="NORDIC")

    # download_nlloc()

    folder_inp = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3"
    out = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/nlloc.cf"
    nll_obj = NLLocObj(folder_inp=folder_inp)
    # nll_obj.write_nlloc_cf(out=out,P_flag=True, S_flag=False,rm_outs=True)
    nll_obj.write_nlloc_cf(out=out,P_flag=False, S_flag=True,rm_outs=False)

    # vel_model = "/home/emmanuel/EDCT/SeisMonitor/data/metadata/vel1d_col.csv"
    # out = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/model.dat"
    # write_1d_vel_model(vel_model,out)

    # vel_model = "/home/emmanuel/EDCT/SeisMonitor/data/metadata/CM.xml"
    # out = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/station.dat"
    # write_station_file(vel_model,out)


# https://github.com/amaggi/waveloc/blob/master/nll/nlloc_sample.in


from obspy.core.inventory.inventory import Inventory
from obspy.io.xseed.parser import Parser
from obspy.core.event import Magnitude as Mag
from obspy.core.event.magnitude import StationMagnitude
from obspy.core.event import Catalog
from obspy.core.event import read_events, Comment
from obspy.geodetics import gps2dist_azimuth
from obspy.signal.invsim import estimate_wood_anderson_amplitude_using_response
from obspy import UTCDateTime
from obspy.core.event.base import CreationInfo,WaveformStreamID,TimeWindow
from obspy.core.event.resourceid import ResourceIdentifier

from obspy.core.event.magnitude import Amplitude
# import mtspec
import numpy as np
import scipy
import math 
import types
paz_wa = {'sensitivity': 2800, 'zeros': [0j], 'gain': 1,
          'poles': [-6.2832 - 4.7124j, -6.2832 + 4.7124j]}

# SED_Ml_params = {"a":0.018,"b":2.17}
Ml_params = {"RSNC":{"a":1.019,"b":0.0016,"r_ref":140},
            "RSNC_by_zone":{"Ml_1":{"a":1.2488,"b":0.0024,"c":-2.05},
                            "Ml_2":{"a":1.0563,"b":0.002,"c":-1.760},
                            "Ml_3":{"a":1.0705,"b":0.0013,"c":-1.531},
                            "Ml_4":{"a":1.2399,"b":0.0015,"c":-2.178},
                            "Ml_5":{"a":0.7096,"b":0.0009,"c":-0.690},
                            }
            }

def get_paz_from_response(seed_id,response,
                            datetime=None):
    if isinstance(response,Parser):
        try:
            paz = response.get_paz(seed_id,datetime)
        except:
            return None

    elif isinstance(response,Inventory):
        try:
            response = response.get_response(seed_id,
                                    datetime)
        
            paz_stage = response.get_paz()
        except:
            return None

        sensitivity = response.instrument_sensitivity.value
        gain = paz_stage.normalization_factor
        poles = paz_stage.poles
        zeros = paz_stage.zeros

        paz = {'poles': poles,
                'zeros': zeros,
                'gain': gain,
                'sensitivity': sensitivity}
    return paz

def get_Ml(ampl,epi_dist,mag_type,zone=None):

    if isinstance(mag_type,str):
        if mag_type not in list(Ml_params.keys()):
            raise Exception(f"{mag_type} not in Ml_params")

        if mag_type.upper() == "RSNC":

            if zone != None:
                
                ml_params = Ml_params["RSNC_by_zone"][f"Ml_{zone}"]

                return math.log10(ampl * 1e9) +  ml_params["a"] * math.log10(epi_dist) +\
                       ml_params["b"] * epi_dist  + ml_params["c"]




            else:
                ml_params = Ml_params["RSNC"]

                k = ml_params["a"]*math.log10(ml_params["r_ref"]/100) +\
                    ml_params["b"]* (ml_params["r_ref"]-100) +3
                    
                return math.log10(ampl * 1e3) + ml_params["a"] * math.log10(epi_dist/ml_params["r_ref"]) +\
                                ml_params["b"] * (epi_dist-ml_params["r_ref"]) + k


    elif isinstance(mag_type,types.LambdaType):
        return mag_type(ampl,epi_dist)

def get_Ml_magparams_by_station(st,response,
                                ev_params,
                                trimmedtime=50,
                                waterlevel=10):
    picktime = ev_params["picktime"]
    latitude = ev_params["latitude"]
    longitude = ev_params["longitude"]

    # if st is None or len(st) != 3:
    if st is None or len(st)==0:
        return (None,None,None)

    for tr in st:
        paz = get_paz_from_response(tr.id,response,
                                tr.stats.starttime)
        if paz == None:
            print(f"\t->No response found: {tr.id}-{tr.stats.starttime}")
            return (None,None,None)

        coords = response.get_coordinates(tr.id,tr.stats.starttime)
        # paz_wa = estimate_wood_anderson_amplitude_using_response(response, amplitude,
        #                                             timespan)
        tr.simulate(paz_remove=paz, 
                    paz_simulate=paz_wa, 
                    water_level=waterlevel)


    st.trim(picktime,picktime+trimmedtime)

    # common_channels_info = list(st._get_common_channels_info().keys())
    # channels = list(map(lambda x: x[-1] ,common_channels_info))
    components = [ tr.stats.channel[-1] for tr in st]
    components = list(set(components))

    if len(components) == 1:
        tr= st.select(component=components[0])[0]
        ampl = max(abs(tr.data))
        tr_id = tr.get_id()
    else:
        if ("N" in components) or ("E" in components):
            if "N" in components:
                tr_n = st.select(component="N")[0]
                tr_n_id = tr_n.get_id()
                ampl_n = max(abs(tr_n.data))
            else:
                ampl_n = None
                tr_n_id = None

            if "E" in components:
                tr_e = st.select(component="E")[0]
                tr_e_id = tr_e.get_id()
                ampl_e = max(abs(tr_e.data))
            else:
                ampl_e = None
                tr_e_id = None

            if ampl_n == None:
                ampl = ampl_e
                tr_id = tr_e_id
            elif ampl_e == None:
                ampl = ampl_n
                tr_id = tr_n_id
            else:
                ampls = [ampl_n,ampl_e]
                trs_id = [tr_n_id,tr_e_id]
                ampl = max(ampls)
                tr_id = trs_id[ampls.index(ampl)]
        elif "Z" in components:
            tr= st.select(component="Z")[0]
            ampl = max(abs(tr.data))
            tr_id = tr.get_id()
        else:
            return (None,None,None)


    sta_lat = coords["latitude"]
    sta_lon = coords["longitude"]

    epi_dist, az, baz = gps2dist_azimuth(latitude, longitude,
                                         sta_lat, sta_lon)
    epi_dist = epi_dist / 1e3

    return ampl,epi_dist,tr_id

def Mw_st_processing(st,response,waterlevel,datetime):
    for trace in st:
        paz = get_paz_from_response(trace.id, response,datetime)

        if paz == None:
            print(f"\t->No response found: {trace.id}-{datetime}")
            return None

        # PAZ in SEED correct to m/s. Add a zero to correct to m.
        paz["zeros"].append(0 + 0j)
        trace.detrend()
        trace.simulate(paz_remove=paz, water_level=waterlevel)
    # st.plot()
    return st

def calculate_source_spectrum(frequencies, omega_0, corner_frequency, Q,
    traveltime):
    """
    After Abercrombie (1995) and Boatwright (1980).

    Abercrombie, R. E. (1995). Earthquake locations using single-station deep
    borehole recordings: Implications for microseismicity on the San Andreas
    fault in southern California. Journal of Geophysical Research, 100,
    24003–24013.

    Boatwright, J. (1980). A spectral theory for circular seismic sources,
    simple estimates of source dimension, dynamic stress drop, and radiated
    energy. Bulletin of the Seismological Society of America, 70(1).

    The used formula is:
        Omega(f) = (Omege(0) * e^(-pi * f * T / Q)) / (1 + (f/f_c)^4) ^ 0.5

    :param frequencies: Input array to perform the calculation on.
    :param omega_0: Low frequency amplitude in [meter x second].
    :param corner_frequency: Corner frequency in [Hz].
    :param Q: Quality factor.
    :param traveltime: Traveltime in [s].
    """
    num = omega_0 * np.exp(-np.pi * frequencies * traveltime / Q)
    denom = (1 + (frequencies / corner_frequency) ** 4) ** 0.5
    return num / denom

def fit_spectrum(spectrum, frequencies, 
    traveltime, initial_omega_0,
    initial_f_c,quality_factor=500):
    """
    Fit a theoretical source spectrum to a measured source spectrum.

    Uses a Levenburg-Marquardt algorithm.

    :param spectrum: The measured source spectrum.
    :param frequencies: The corresponding frequencies.
    :para traveltime: Event traveltime in [s].
    :param initial_omega_0: Initial guess for Omega_0.
    :param initial_f_c: Initial guess for the corner frequency.


    :returns: Best fits and standard deviations.
        (Omega_0, f_c, Omega_0_std, f_c_std)
        Returns None, if the fit failed.
    """
    def f(frequencies, omega_0, f_c):
        return calculate_source_spectrum(frequencies, omega_0, f_c,
                quality_factor, traveltime)
    popt, pcov = scipy.optimize.curve_fit(f, frequencies, spectrum, \
        p0=list([initial_omega_0, initial_f_c]), maxfev=100000)
    if popt is None:
        return None
    return popt[0], popt[1], pcov[0, 0], pcov[1, 1]

def get_M0_magnitude_by_pick(st,picktime,traveltime,
                                phasehint,
                                physparams,
                                procparams):


    if st is None or len(st) != 3:
        return (None,None)

    if phasehint.upper() == "P":
        velocity = physparams.vp
        radiation_pattern = physparams.p_radiation_pattern
    elif phasehint.upper() == "S":
        velocity = physparams.vs
        radiation_pattern = physparams.s_radiation_pattern
    else:
        return (None,None)

    distance = traveltime * velocity
    if distance <= 0.0:
        return None

    omegas = []
    corner_freqs = []
    for trace in st:
        pick_index = int(round((picktime - trace.stats.starttime) / \
                                trace.stats.delta))
        # Choose date window 0.5 seconds before and 1 second after pick.
        data_window = trace.data[pick_index - \
                    int(procparams.time_before_pick * trace.stats.sampling_rate): \
                    pick_index + int(procparams.time_after_pick * trace.stats.sampling_rate)]

        spec, freq = mtspec.mtspec(data_window, trace.stats.delta, 2)
        try:
            fit = fit_spectrum(spec, freq, traveltime,
                    spec.max(), 10.0)
        except:
            continue

        if fit is None:
            continue
        Omega_0, f_c, err, _ = fit
        Omega_0 = np.sqrt(Omega_0)
        omegas.append(Omega_0)
        corner_freqs.append(f_c)

    if omegas:

        M_0 = 4.0 * np.pi * physparams.density * velocity ** 3 * distance * \
            np.sqrt(omegas[0] ** 2 + omegas[1] ** 2 + omegas[2] ** 2) / \
            radiation_pattern

        return (M_0,corner_freqs)
    else:
        return (None,None)

def write_magsta_values(value,
                        uncertainty,
                        mag_type,
                        origin_id = ResourceIdentifier(),
                        amplitude_id = ResourceIdentifier(),
                        method_id = ResourceIdentifier(),
                        waveform_id = WaveformStreamID(),
                        agency=None ):
    stamag = StationMagnitude()
    stamag.resource_id = ResourceIdentifier()
    stamag.mag = value
    stamag.mag_errors.uncertainty = uncertainty
    stamag.station_magnitude_type = mag_type
    stamag.origin_id = origin_id
    stamag.method_id = method_id
    stamag.amplitude_id = amplitude_id 
    stamag.waveform_id = waveform_id
    stamag.creation_info = CreationInfo(agency_id=agency,
										agency_uri=ResourceIdentifier(id=agency),
										author="SeisMonitor",
										author_uri=ResourceIdentifier(id="SeisMonitor"),
										creation_time=UTCDateTime.now())
    return stamag

def write_amplitude_values(value,amp_type="A",category="duration",
                           unit="m/s",
                           method_id = "https://docs.obspy.org/tutorial/advanced_exercise/advanced_exercise_solution_3b.html",
                           time_window=TimeWindow(),
                           pick_id = ResourceIdentifier(),
                           waveform_id = WaveformStreamID(),
                           magnitude_hint = "M",
                           evaluation_mode = "automatic",
                           evaluation_status = "preliminary",
                           agency=None ):
    amp = Amplitude()
    amp.resource_id = ResourceIdentifier()
    amp.generic_amplitude = value
    amp.type = amp_type
    amp.category = category
    amp.unit = unit
    amp.method_id = method_id
    amp.time_window = time_window
    amp.pick_id = pick_id 
    amp.waveform_id = waveform_id
    amp.magnitude_hint  = magnitude_hint
    amp.evaluation_mode = evaluation_mode
    amp.evaluation_status = evaluation_status
    amp.creation_info = CreationInfo(agency_id=agency,
										agency_uri=ResourceIdentifier(id=agency),
										author="SeisMonitor",
										author_uri=ResourceIdentifier(id="SeisMonitor"),
										creation_time=UTCDateTime.now())
    return amp

def write_magnitude_values(value,uncertainty,station_count,mag_type,
                           evaluation_mode = "automatic",
                           evaluation_status = "preliminary",
                           method_id=ResourceIdentifier(),
                           agency=None,
                           origin_id=ResourceIdentifier(),
                           comments=None):
    mag = Mag()
    mag.mag = value
    mag.mag_errors.uncertainty = uncertainty
    mag.magnitude_type = mag_type
    mag.origin_id = origin_id
    mag.station_count = station_count
    mag.evaluation_mode = evaluation_mode 
    mag.evaluation_status = evaluation_status
    mag.method_id = method_id
    mag.comments.append(Comment( \
        "Magnitude=%e Nm; uncertainty=%e" % (value,
        uncertainty)))
    if comments != None:
        mag.comments.append(Comment(comments))
    mag.creation_info = CreationInfo(agency_id=agency,
										agency_uri=ResourceIdentifier(id=agency),
										author="SeisMonitor",
										author_uri=ResourceIdentifier(id="SeisMonitor"),
										creation_time=UTCDateTime.now())
    
    return mag


import os 
import math
from obspy import read_events
from obspy.clients.fdsn import Client as FDSNClient
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.event.origin import OriginUncertainty
from obspy.core.event.event import Event
from SeisMonitor.monitor.seismonitor import SeisMonitor
from SeisMonitor.monitor.picker import ai as ai_picker
from SeisMonitor.monitor.downloader import utils as dut
from SeisMonitor.monitor.associator import ai as ai_asso
from SeisMonitor.monitor.locator.nlloc.nlloc import NLLoc
from SeisMonitor.monitor.locator import utils as lut
from SeisMonitor.core.objects import WaveformRestrictions,Provider
from SeisMonitor.monitor.downloader.utils import get_merged_inv_and_json

out = "/home/emmanuel/E_ColSeismicity/ColSeismicity/test"

sgc_client = FDSNClient('http://sismo.sgc.gov.co:8080')
sgc_rest = WaveformRestrictions(network="CM",
                    station="URMC,VILL,PRA,CLEJA",
                    location="*",
                    channel="*",
                    starttime=UTCDateTime("2019-12-24T19:00:00.000000Z"),
                    endtime=UTCDateTime("2019-12-25T01:00:00.000000Z"),
                    location_preferences=["","00","20","10","40"],
                    channel_preferences=["HH","BH","EH","HN","HL"],
                    filter_networks=[], 
                    filter_stations=[],
                    filter_domain= [-83.101,-64.549,-2.229,14.945],
                    )
sgc_client = FDSNClient('http://sismo.sgc.gov.co:8080')
sgc_provider = Provider(sgc_client,sgc_rest)

seismo = SeisMonitor(providers = [sgc_provider],
                    chunklength_in_sec=7200,
                    out_folder = out)

seismo.add_downloader(picker_args= {"batch_size":100,"overlap":0.3,"length":60})

dataset = os.path.join(os.path.dirname(os.path.dirname(__file__)),"data")
eqt_model = os.path.join(dataset,"models",'EqT_model.h5')
seismo.add_picker(
                  pickers={
                            "EQTransformer":ai_picker.EQTransformerObj(
                                            model_path = eqt_model,
                                            n_processor = 32,
                                            overlap = 0.3,
                                            detection_threshold =0.1,
                                            P_threshold = 0.01,
                                            S_threshold = 0.01,
                                            batch_size = 100,
                                            number_of_plots = 0,
                                            plot_mode = 1,
                                            rm_downloads=True ) 
                            }
                )
# seismo.add_associator(input=["EQTransformer"],
#                         associators={
#                         "GaMMA":ai_asso.GaMMAObj(
#                                             [-85, -68,-2, 15,0, 180],
#                                             "EPSG:3116",
#                                             use_amplitude = False,
#                                             use_dbscan=False,
#                                             max_sigma11=5.0,
#                                             calculate_amp=False)
#                         }
#                         )

# dataset = os.path.join(os.path.dirname(os.path.dirname(__file__)),"data")
# vel_path = os.path.join(dataset,"metadata","vel1d_col.csv")
# vel_model = lut.VelModel(vel_path)
# inv,_,_,_ = dut.get_merged_inv_and_json(seismo.providers)
# stations = lut.Stations(inv)

# nlloc = NLLoc(
#         agency="SeisMonitor",
#         region = [-85, -68,0, 15,-5, 205],
#         vel_model = vel_model,
#         stations = stations,
#         delta_in_km = 2.5,
#         tmp_folder="/home/emmanuel/NLLoc_grid/NLLoc_grid" ### CHANGE PATH TO YOUR OWN PATH AND ALSO TAKE IN MIND THAT CONSUME DISK
#         )

# seismo.add_locator(input={"associations":("GaMMA","EQTransformer")},
#                     locators={
#                         "NLLOC":nlloc}
#                         )


# ml_params = {"a":1.019,"b":0.0016,"r_ref":140} #ojeda
# k = ml_params["a"]*math.log10(ml_params["r_ref"]/100) +\
#                     ml_params["b"]* (ml_params["r_ref"]-100) +3
# Ml = lambda ampl,epi_dist : math.log10(ampl * 1e3) + ml_params["a"] * math.log10(epi_dist/ml_params["r_ref"]) +\
#                                 ml_params["b"] * (epi_dist-ml_params["r_ref"]) + k

# seismo.add_magnitude(input={"locations":("NLLOC","GaMMA/EQTransformer")},
#                     magnitudes={
#                         "Ml":{"mag_type":Ml,
#                                 "trimmedtime":5,
#                                 "out_format":"SC3ML"}}
#                                 )
seismo.run()